#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "auto_uids.h"
#include "substdio.h"
#include "strerr.h"
#include "error.h"
#include "open.h"
#include "readwrite.h"
#include "exit.h"

extern void hier();

#define FATAL "install: fatal: "

int fdsourcedir = -1;

typedef struct {
    const char* qu_name;
    int qu_id;
} qmail_uid_t;

qmail_uid_t *qmail_uids = NULL;
qmail_uid_t *qmail_gids = NULL;

void init_ids(qmail_uid_t** dst, const qmail_uid_t* src, size_t size)
{
  int i;

  (*dst) = malloc(size*sizeof(qmail_uid_t));
  for (i=0; i < size; i++) {
    if (src[i].qu_name != NULL) {
      (*dst)[i].qu_name = strdup(src[i].qu_name);
    } else {
      (*dst)[i].qu_name = NULL;
    }
    (*dst)[i].qu_id = src[i].qu_id;
  }
}

static const char* get_name(const qmail_uid_t* table, int id)
{
  const qmail_uid_t* iter;

  for (iter = table; iter->qu_name != NULL; iter++) {
    if (iter->qu_id == id)
      return iter->qu_name;
  }

  fprintf(stderr, "No user/group with id=%d\n", id);
  exit(1);
}

static FILE* out = NULL;
void write_log(int uid, int gid, const char* path, const char* dir, const char* file)
{
  const char* user;
  const char* group;
  static char* prefix = NULL;
  static size_t prefix_len = 0;
  const qmail_uid_t uids[] = {
    {"alias", auto_uida},
    {"qmaild", auto_uidd},
    {"qmaill", auto_uidl},
    {"root", auto_uido},
    {"qmailp", auto_uidp},
    {"qmailq", auto_uidq},
    {"qmailr", auto_uidr},
    {"qmails", auto_uids},
    {NULL, 0}
  };
  const qmail_uid_t gids[] = {
    {"0", 0},
    {"qmail", auto_gidq},
    {"nofiles", auto_gidn},
    {NULL, 0}
  };

  if (uid == 0 && gid == 0)
    return;

  if (qmail_uids == NULL) {
    init_ids(&qmail_uids, uids, sizeof(uids)/sizeof(qmail_uid_t));
    init_ids(&qmail_gids, gids, sizeof(gids)/sizeof(qmail_uid_t));
  }

  if (out == NULL) {
    char* out_filename = getenv("FILE_LIST");
    if (out_filename == NULL) {
      out = stderr;
    } else {
      out = fopen(out_filename, "w");
      if (out == NULL) {
        perror("Cannot open $FILE_LIST for append");
        fprintf(stderr, "File: %s\n", out_filename);
        exit(1);
      }
    }
  }

  if (prefix == NULL) {
    prefix = getenv("FILE_LIST_PREFIX");
    if (prefix == NULL) {
	prefix = "";
    } else {
	prefix_len = strlen(prefix);
    }
  }

  user = get_name(qmail_uids, uid);
  group = get_name(qmail_gids, gid);
  if (strncmp(path, prefix, prefix_len) == 0) {
    path += prefix_len;
  }
  if (file == NULL) {
    fprintf(out, "%s,%s,%s\n", user, group, path);
  } else if (dir == NULL) {
    fprintf(out, "%s,%s,%s/%s\n", user, group, path, file);
  } else {
    fprintf(out, "%s,%s,%s/%s/%s\n", user, group, path, dir, file);
  }
}

void fix_gid(int* gid_p, int mode)
{
  // if perms for group == perms for others, and no setguid, set group to root
  if (((mode&S_IRWXO) == ((mode&S_IRWXG)>>3)) && (mode&S_ISGID) == 0)
    *gid_p = 0;
}

void h(home,uid,gid,mode)
char *home;
int uid;
int gid;
int mode;
{
  fix_gid(&gid, mode);
  if (mkdir(home,0700) == -1)
    if (errno != error_exist)
      strerr_die4sys(111,FATAL,"unable to mkdir ",home,": ");
  write_log(uid, gid, home, NULL, NULL);
  if (chmod(home,mode) == -1)
    strerr_die4sys(111,FATAL,"unable to chmod ",home,": ");
}

void d(home,subdir,uid,gid,mode)
char *home;
char *subdir;
int uid;
int gid;
int mode;
{
  fix_gid(&gid, mode);
  if (chdir(home) == -1)
    strerr_die4sys(111,FATAL,"unable to switch to ",home,": ");
  if (mkdir(subdir,0700) == -1)
    if (errno != error_exist)
      strerr_die6sys(111,FATAL,"unable to mkdir ",home,"/",subdir,": ");
  write_log(uid, gid, home, NULL, subdir);
  if (chmod(subdir,mode) == -1)
    strerr_die6sys(111,FATAL,"unable to chmod ",home,"/",subdir,": ");
}

void p(home,fifo,uid,gid,mode)
char *home;
char *fifo;
int uid;
int gid;
int mode;
{
  fix_gid(&gid, mode);
  if (chdir(home) == -1)
    strerr_die4sys(111,FATAL,"unable to switch to ",home,": ");
  if (fifo_make(fifo,0700) == -1)
    if (errno != error_exist)
      strerr_die6sys(111,FATAL,"unable to mkfifo ",home,"/",fifo,": ");
  write_log(uid, gid, home, NULL, fifo);
  if (chmod(fifo,mode) == -1)
    strerr_die6sys(111,FATAL,"unable to chmod ",home,"/",fifo,": ");
}

char inbuf[SUBSTDIO_INSIZE];
char outbuf[SUBSTDIO_OUTSIZE];
substdio ssin;
substdio ssout;

void c(home,subdir,file,uid,gid,mode)
char *home;
char *subdir;
char *file;
int uid;
int gid;
int mode;
{
  int fdin;
  int fdout;

  fix_gid(&gid, mode);

  if (fchdir(fdsourcedir) == -1)
    strerr_die2sys(111,FATAL,"unable to switch back to source directory: ");

  fdin = open_read(file);
  if (fdin == -1)
    strerr_die4sys(111,FATAL,"unable to read ",file,": ");
  substdio_fdbuf(&ssin,read,fdin,inbuf,sizeof inbuf);

  if (chdir(home) == -1)
    strerr_die4sys(111,FATAL,"unable to switch to ",home,": ");
  if (chdir(subdir) == -1)
    strerr_die6sys(111,FATAL,"unable to switch to ",home,"/",subdir,": ");

  fdout = open_trunc(file);
  if (fdout == -1)
    strerr_die6sys(111,FATAL,"unable to write .../",subdir,"/",file,": ");
  substdio_fdbuf(&ssout,write,fdout,outbuf,sizeof outbuf);

  switch(substdio_copy(&ssout,&ssin)) {
    case -2:
      strerr_die4sys(111,FATAL,"unable to read ",file,": ");
    case -3:
      strerr_die6sys(111,FATAL,"unable to write .../",subdir,"/",file,": ");
  }

  close(fdin);
  if (substdio_flush(&ssout) == -1)
    strerr_die6sys(111,FATAL,"unable to write .../",subdir,"/",file,": ");
  if (fsync(fdout) == -1)
    strerr_die6sys(111,FATAL,"unable to write .../",subdir,"/",file,": ");
  if (close(fdout) == -1) /* NFS silliness */
    strerr_die6sys(111,FATAL,"unable to write .../",subdir,"/",file,": ");

  write_log(uid, gid, home, subdir, file);
  if (chmod(file,mode) == -1)
    strerr_die6sys(111,FATAL,"unable to chmod .../",subdir,"/",file,": ");
}

void z(home,file,len,uid,gid,mode)
char *home;
char *file;
int len;
int uid;
int gid;
int mode;
{
  int fdout;

  fix_gid(&gid, mode);

  if (chdir(home) == -1)
    strerr_die4sys(111,FATAL,"unable to switch to ",home,": ");

  fdout = open_trunc(file);
  if (fdout == -1)
    strerr_die6sys(111,FATAL,"unable to write ",home,"/",file,": ");
  substdio_fdbuf(&ssout,write,fdout,outbuf,sizeof outbuf);

  while (len-- > 0)
    if (substdio_put(&ssout,"",1) == -1)
      strerr_die6sys(111,FATAL,"unable to write ",home,"/",file,": ");

  if (substdio_flush(&ssout) == -1)
    strerr_die6sys(111,FATAL,"unable to write ",home,"/",file,": ");
  if (fsync(fdout) == -1)
    strerr_die6sys(111,FATAL,"unable to write ",home,"/",file,": ");
  if (close(fdout) == -1) /* NFS silliness */
    strerr_die6sys(111,FATAL,"unable to write ",home,"/",file,": ");

  write_log(uid, gid, home, NULL, file);
  if (chmod(file,mode) == -1)
    strerr_die6sys(111,FATAL,"unable to chmod ",home,"/",file,": ");
}

int main()
{
  fdsourcedir = open_read(".");
  if (fdsourcedir == -1)
    strerr_die2sys(111,FATAL,"unable to open current directory: ");

  umask(077);
  hier();
  fclose(out);
  _exit(0);
}
